#ifndef DATA_H
#define DATA_H

//#define	VOLT5	/* version for 5V supply, slightly different pinning */
#define	USE_IR
#define USE_SERVO
#define USE_ROBO_EXT
#define USE_DISTANCE
//#define USE_STEPPER

#if defined(USE_STEPPER) //modules that also need I2C
#define USE_I2C
#endif

#include <stdint.h>
typedef uint8_t UINT8;
typedef uint16_t UINT16;
typedef uint8_t BOOL8;
typedef int16_t INT16;
typedef int32_t INT32;
typedef uint16_t BOOL16;
typedef uint8_t UCHAR8;
typedef uint32_t UINT32;

typedef enum {false, true} _bool;
//#include "ROBO_TX_FW.h" //"../PC_programming_RoboTXC_V1-2_11_Dec_2009/PC_Programming_RoboTXC/Demo_Static_Lib_C/Inc/"
#include "ROBO_TX_FW 1_24.h" //for TX FW 1.24

typedef struct
{ UINT16 start;
  union //the message format is HL, the native format is LH, receiver routine swaps HL to LH, send routine writes as HL
  { UINT16 bytes;
    struct
    { UINT8 bytesH, bytesL;
	}/*anonymous*/; //as_bytes;
  };//winavr allows anonymous structs and unions
  UINT32 snd, rec;
  UINT16 trans, session;
  UINT32 cmd, structs, ta_id;//ta_id is normally not considered part of the header but part of the payload, it is repeated for each struct
} header;


typedef struct
{ UINT16	chksum;
  UCHAR8	etx;
} trailer;

typedef struct
{ header hdr;
  union
  { struct { TA_OUTPUT output; trailer trl;} cmd002;//set outputs
    struct { TA_CONFIG config; trailer trl;} cmd005;//initialize
	struct {                   trailer trl;} cmd006;//extension request
	struct { TA_DISPLAY disp;  trailer trl;} cmd008;//display message
	struct { char name[20];    trailer trl;} cmd009;//change name, 20 = DEV_NAME_LEN+1+3padding
	struct { TA_INPUT  input;  trailer trl;} cmd102;//reply to cmd002
    struct {                   trailer trl;} cmd105;//reply to cmd005
	struct { TA_INFO   info;   trailer trl;} cmd106;//reply to cmd006
    struct {                   trailer trl;} cmd108;//reply to cmd008
    struct {                   trailer trl;} cmd109;//reply to cmd009
    struct {                   trailer trl;} empty; //empty reply (cmd108,109)
  } body;
} message;

#define MAXEXT	4	//more extensions are technically possible but the 10ms update time has to be enlarged
#define SLAVES	8

struct _to_ext
{ UCHAR8 res;
  UCHAR8 motors;
  UINT32 pwm;
};

struct _from_ext
{ UCHAR8 inputs;
  UINT8 ax;
  UINT8 az;
  UINT8 av;
  UINT8 high;
  UCHAR8 res;
};

struct _conf
{ char motor; //4 lsb only, if 1 output is full H bridge (motor), if 0 output is 2 half bridges (lamps)
  char uni;   //8 bits, if 1 input is digital, if 0 input is analog
  char dist;  //8 bits, if 1 input is distance
  char resist;//8 bits, if 1 input is in resistance mode
};

extern struct _to_ext to_ext[MAXEXT];
extern struct _from_ext from_ext[MAXEXT];
extern struct _conf config[SLAVES];
extern message msg;
extern unsigned char thisindex;
extern char eeprom_busy;
extern unsigned char bytemap[];
void store_index();
void eeprom_write_block_nonblocking(char *src, char *dst, int size);
//short int find_ad(uint32_t);

//#define DEBUG		//define DEBUG in the unit to be debugged
//pin numbers are wrong, they are 15 and 16 now
#ifdef DEBUG
#define SETPIN11	  asm("sbi	%0, %1"::"I" (_SFR_IO_ADDR(PORTB)),"I" (PB1));//debug11(1);
#define CLRPIN11	  asm("cbi	%0, %1"::"I" (_SFR_IO_ADDR(PORTB)),"I" (PB1));//debug11(0);
#define SETPIN12	  asm("sbi	%0, %1"::"I" (_SFR_IO_ADDR(PORTB)),"I" (PB2));//debug12(1);
#define CLRPIN12	  asm("cbi	%0, %1"::"I" (_SFR_IO_ADDR(PORTB)),"I" (PB2));//debug12(0);
#else
#define SETPIN11
#define CLRPIN11
#define SETPIN12
#define CLRPIN12
#endif	

#ifdef VOLT5
#define LEDON	  asm("sbi	%0, %1"::"I" (_SFR_IO_ADDR(PORTD)),"I" (PD4));
#define LEDOFF	  asm("cbi	%0, %1"::"I" (_SFR_IO_ADDR(PORTD)),"I" (PD4));
#endif

#define BYTES()	((msg.hdr.bytesH<<8)|msg.hdr.bytesL)

#endif
