//#define DEBUG	//PB1 switches in interrupt and timeout
#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include "data.h"
#include "robo_ext.h"


#define ADPORT	PORTD	//port for 3 bit io ext. address
#define ADBIT	PD5		//PD5-7
#define ADDIR	DDRD


#define SSN		PB2
#define MOSI	PB3
#define MISO	PB4	
#define SCK		PB5

#ifdef USE_ROBO_EXT

char temp_from_ext[sizeof(struct _from_ext)], temp_to_ext[sizeof(struct _to_ext)];
short int bytes = 0, extension = 7;
short int ack_timeout = 0;

void set_address(char ad)
{
#ifdef OPENDRAIN //output is never actively driven high
#ifdef EXTPULLUP //no pullup, rely on external pullup to 3.3V
#pragma warn "address bus needs external pullups"
  //ADPORT = (ADPORT & ~(7<<ADBIT));  //never changes, put in init
  ADDIR = (ADDIR | (7<<ADBIT)) & ~(ad<<ADBIT);
#else  //use internal pull-up
  ADDIR = (ADDIR & ~(7<<ADBIT)); //make Hi-Z to avoid glitch to VCC
  ADPORT = (ADPORT & ~(7<<ADBIT)) | (ad<<ADBIT); //set port
  ADDIR = (ADDIR | (7<<ADBIT)) & ~(ad<<ADBIT); //set direction to inverse of port
#endif
#else //normal digital output
#pragma warn "actively driving OD lines high"
  ADPORT = (ADPORT & ~(7<<ADBIT)) | (ad<<ADBIT);
#endif
}

void init_extension()
{ //init SPI, ca. 100kHz clock, master, no interrupt
  DDRB |= _BV(MOSI)|_BV(SCK); //set MOSI and SCK as output
  SPCR = SPI_FLAGS |_BV(SPIE);
  //init ack and address ports
  ACKDIR &= ~_BV(ACKBIT); //set ack as input
  ACKPORT |= _BV(ACKBIT); //enable internal pullup
#ifdef OPENDRAIN
#ifdef EXTPULLUP
  ADPORT &= ~(7<<ADBIT); //set address bits low for active pull-down 
#else //no ext. pullup
  ADPORT |= 7<<ADBIT; //set address to 7 
#endif  //EXTPULLUP
  ADDIR &= ~(7<<ADBIT);  //set 3 consecutive bits as input, pullup will make them high
#else//no open drain
  ADPORT |= 7<<ADBIT; //set address to 7 
  ADDIR |= 7<<ADBIT;  //set 3 consecutive bits as output
#endif//OPENDRAIN
  PORTB &= ~_BV(MOSI); //set MOSI low
  PORTB |= _BV(SCK)|_BV(MISO);	//set SCK high and enable pull-up on MISO
  //init ack int
#if ACKPIN == INT0
  EICRA = (EICRA & ~_BV(ISC00)) | _BV(ISC01); //int0 falling edge
  EIMSK |= _BV(INT0);
#elif ACKPIN == INT1
  EICRA = (EICRA & ~_BV(ISC10)) | _BV(ISC11); //int1 falling edge
  EIMSK |= _BV(INT1);
#else
  PCICR |= _BV(PCIE);
  PCMASK |= _BV(ACKBIT);
#endif
}

#if 1
//use both ACK and SPI interrupt

static char state = 0;

void read_extensions()
{ extension = 0;
  //set first timeout
  ack_timeout = 20; //ticks of 100us = 2ms
  //output 'extension' as 'address' and wait for an ack int
  cli(); //this routine (although timer interrupt) can be interrupted by rx/tx which use the same port
  //ADPORT = (ADPORT & ~(7<<ADBIT)) | (extension<<ADBIT);
  set_address(extension);
  SPCR |= _BV(SPIE);
  sei();
}

//there is a race between the ack and the spi interrupt but spi should come first
//data modified by more than 1 interrupt routine:
// state (spi, ack)
// bytes (spi, ack)
// ack_timeout (timer, ack)
// extensions (timer, timeout, ack)
// nrofext (timeout, ack)
// temp_from_ext (spi, ack)

ISR(SPI_STC_vect)  //should only happen in state 1
{ temp_from_ext[bytes] = SPDR;
  if (bytes < sizeof(struct _from_ext)-1)
  { bytes++;
    //SPDR = temp_to_ext[bytes]; //postpone send till after ack
	state = 3;
  }
  else
    state = 2;
}

ISR(ACKINT_vect, ISR_NOBLOCK)//should only trigger on change in ack
{ //SETPIN11;
  cli();
  switch (state) //make operations on 'state' and 'bytes' atomic
  { case 0: //ack goes low in response to change of address 
 	        state = 1;
	        bytes = 0;
			sei();
            memcpy(temp_to_ext, &to_ext[extension], sizeof(struct _to_ext)); //copy the data for extension
            SPDR = temp_to_ext[0]; //send the first byte
        	break;
    case 1: //reading data, handled by SPI
	        //here we should deal with the case that ack precedes spi
			//2 options: 1: write the new data before the old data is read and rely on spi int to take care of reading, danger: small risk of overrun
			//           2: flag the spi that ack happened and do the write in spi int, danger: sensitive to glitches
	        sei();
	        break;
    case 2: //last byte transfered
			state = 0;
	        sei();
            //copy buffer to struct for this extension
            memcpy(&from_ext[extension], temp_from_ext, sizeof(struct _from_ext));
	        extension++; 
            if (extension>=MAXEXT)
            { extension = 7; //disable, interrupts should stop because no more ack's
              nrofext = MAXEXT;
              ack_timeout = 0; //no timeouts anymore, all extensions treated
            }
            else
              ack_timeout = 20; //ticks of 100us
            //output 'extension' as 'address'
            cli(); //this routine (although interrupt) can be interrupted by rx/tx which use the same port
            //ADPORT = (ADPORT & ~(7<<ADBIT)) | (extension<<ADBIT);
			set_address(extension);
	        sei();
			break;
    case 3: //byte received, send next
            SPDR = temp_to_ext[bytes];
			state = 1;
			sei();
			break;
  }
  CLRPIN11;
}

#else
//use only ACK interrupt (sensitive to glitches)

void read_extensions()
{ extension = 0;
  bytes = 0;
  //set first timeout
  ack_timeout = 20; //ticks of 100us = 2ms
  //output 'extension' as 'address' and wait for an ack int
  memcpy(temp_to_ext, to_ext, sizeof(struct _to_ext)); //copy the data for extension[0]
  cli(); //this routine (although timer interrupt) can be interrupted by rx/tx which use the same port
  set_address(extension);
  sei();
}

ISR(ACKINT_vect, ISR_NOBLOCK)//should only trigger on change in ack
{ //SETPIN11;
#if ACKPIN != INT0 && ACKPIN != INT1
  if (ACKPORT & _BV(ACKBIT)) //test whether ack went high
  { 
    return;
  }
#endif
  if (bytes >= sizeof(struct _from_ext))
  { //read last byte
//SETPIN11;
//    while (!(SPSR & _BV(SPIF))) SETPIN11;
    if (SPSR & _BV(SPIF)) //was a character received?
	{
      temp_from_ext[sizeof(struct _from_ext)-1] = SPDR;
      bytes = 0;
      //copy buffer to struct for this extension
      memcpy(&from_ext[extension++], temp_from_ext, sizeof(struct _from_ext));
      if (extension>=MAXEXT)
      { extension = 7; //disable, interrupts should stop because no more ack's
        ack_timeout = 0; //no timeouts anymore, all extensions treated
        nrofext = MAXEXT;
      }
      else
      { //copy output buffer for next extension
        memcpy(temp_to_ext, &to_ext[extension], sizeof(struct _to_ext));
        //set timeout
        ack_timeout = 20; //ticks of 100us
      }
      //output 'extension' as 'address'
      cli(); //this routine (although interrupt) can be interrupted by rx/tx which use the same port
      set_address(extension);
 	  sei();
	}
	else
	{ //probably a glitch
	  SETPIN11;
    }
  }
  else
  { if (SPSR & _BV(SPIF))
    {
      if (bytes>0)//should be true otherwise SFIF would be false
      { //read byte
	    //while (!(SPSR & _BV(SPIF))) SETPIN11;
        temp_from_ext[bytes-1] = SPDR;
      }
//	  if (SPSR & _BV(WCOL)) SETPIN11;
      // send next byte
      SPDR = temp_to_ext[bytes];
      bytes++;
    }
	else if (bytes == 0) //first time
	{ SPDR = temp_to_ext[bytes];
      bytes++;
	}
	else
	{ //probably a glitch
	  SETPIN11;
    }
  }
  CLRPIN11;
}
#endif

void extension_timeout()
{ //SETPIN11;
  nrofext = extension;
  extension = 7;
  cli(); //this routine (although timer interrupt) can be interrupted by rx/tx which use the same port
  set_address(extension);
  sei();
  //CLRPIN11;
}

#endif
